/*
 *  dialog - Display simple dialog boxes from shell scripts
 *
 *  Originally by Savio Lam (lam836@cs.cuhk.hk)
 *  Completely reworked by Alessandro Rubini (rubini@ipvvis.unipv.it).
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "dialog.h"
#include "dialogInt.h"
#include <errno.h>

#ifdef LOCALE
#include <locale.h>
#endif

static void Usage (const char *name);

static int opt_separate = 0;

typedef int (dialogFn)(Dialog_desc *desc);
 
struct Dialog_desc desc; /* zero filled, 'cause it's static */

/*
 * The following eases parsing of the cmdline.
 */

enum Optiontype {OPT_FLG=1, OPT_STR, OPT_INT, OPT_JMP};

struct Option {
    char name;
    char *longname;
    int type;
    void *ptr;
};    

/*
 * Variables to keep option values
 */
static short opt_clear=0, opt_one=0;
static char *opt_opts, *opt_file;

/*
 * Finally, the array of options.
 */

static struct Option options[] =
{

/* old ones */

  { 'T', "-title",           OPT_STR, &(desc.title)},
  { 'B', "-backtitle",       OPT_STR, &(desc.backtitle)},
  { 'S', "-separate-output", OPT_FLG, &opt_separate},
  { 'C', "-clear",           OPT_FLG, &opt_clear},

/* new ones */

  { 'o', "-options",         OPT_STR, &opt_opts},
  { 's', "-select",          OPT_INT, &(desc.selected)},
  { 'x', "-extra",           OPT_INT, &(desc.xtra)},
  { 'f', "-file",            OPT_STR, &opt_file},
  { '1', "-one-column",      OPT_FLG, &opt_one},

/* jumpers (types) */

  { 'y', "-yesno",           OPT_JMP, dialog_simple},
  { 'k', "-msgbox",          OPT_JMP, dialog_simple},
  { 'i', "-infobox",         OPT_JMP, dialog_simple},
  { 't', "-textbox",         OPT_JMP, dialog_textbox},
  { 'm', "-menu",            OPT_JMP, dialog_list},
  { 'c', "-checklist",       OPT_JMP, dialog_list},
  { 'r', "-radiolist",       OPT_JMP, dialog_list},
  { 'q', "-inputbox",        OPT_JMP, dialog_inputbox},
  { 'g', "-gauge",           OPT_JMP, dialog_gauge},

  {  0, NULL, 0, NULL}
};

static struct Option *optPtr;

int
main (int argc, char **argv)
{
    int retval;
    const char *prgname=argv[0];
    Dialog_list *lPtr=NULL;
    FILE *infile=NULL, *outfile=stderr;

#ifdef LOCALE
    (void) setlocale (LC_ALL, "");
#endif

    if (argc < 2) {
	Usage (prgname);
	exit (DIALOG_ERROR);
    }

    /*
     * Parse options.
     */
    while (argc>1 && argv[1][0]=='-') {
        argv++, argc--; /* eat it */
	for (optPtr=options; optPtr->name; optPtr++) {
	    if (argv[0][1] != optPtr->name && strcmp(*argv+1,optPtr->longname))
	        continue;

	    /* found */
	    switch(optPtr->type) {
	        case OPT_FLG:
	            (*(short *)(optPtr->ptr))++; break;

		case OPT_STR:
		    if (argc<2) { Usage(prgname); exit(DIALOG_ERROR); }
		    *(char **)(optPtr->ptr)=argv[1];
		    argv++; argc--; break;

		case OPT_INT:
		    if (argc<2) { Usage(prgname); exit(DIALOG_ERROR); }
		    *(short *)(optPtr->ptr)=(short)atoi(argv[1]);
		    argv++; argc--; break;

		case OPT_JMP:
		    desc.type=optPtr->name;
		    /* This is the last option -- break the loop */
		    argv--; argc++; argv[1][0]='!'; /* dirty! */
		    break;

		default:
		    fprintf(stderr,"%s: can't happen\n",prgname);
		    exit(DIALOG_ERROR);
	    }
	    break; /* next arg */
	}
	if (optPtr->name == 0)
	    break; /* not found */
    }

    if (argc<3) {
        Usage (prgname);
	exit(DIALOG_ERROR);
    }
    if (optPtr->name==0 || argv[1][0]!='!') {
	fprintf(stderr,"%s: Error parsing \"%s\"\n",prgname,argv[1]);
	exit(DIALOG_ERROR);
    }
    argv++,argc--; /* now, 0 is the flavour specs */

    /*
     * Process common options and size specifications
     */

    if (opt_opts)
      dialog_defaultbuttons(&desc,opt_opts); /* opt_opts is now unused */

    desc.prompt=argv[1]; argv++,argc--;

    if (argc>1 && sscanf(argv[1],"%hi%s",&desc.c_hei,opt_opts)==1) {
	argv++,argc--;
    } else {
	desc.c_hei=0;
    }

    if (argc>1 && sscanf(argv[1],"%hi%s",&desc.c_wid,opt_opts)==1) {
	argv++,argc--;
    } else {
	desc.c_wid=0;
    }


    /*
     * Now fix files: stdin may be input, stdout may be output. Those which
     * are not a tty are reopened, and stderr is used for output only if
     * both of them are ttys (compatibility mode). If stderr is a tty, use
     * its ttyname() instead of /dev/tty, so Gpm_Open will know which console
     * we are running on.
     */

    {
	char *tty="/dev/tty";
	if (isatty(fileno(stderr)))
	  tty=ttyname(fileno(stderr));

	if (!isatty(fileno(stdin))) {
	    int fd=dup(fileno(stdin));
	    infile=fdopen(fd,"r");
	    if (!freopen(tty,"r",stdin)) {
		fprintf(stderr,"%s: %s: %s\n",prgname,tty,strerror(errno));
		exit(DIALOG_ERROR);
	    }
	}
	
	/* but "-f" gets priority over stdin */
	if (opt_file) {
	    if (infile) fclose(infile);
	    if (!(infile=fopen(opt_file,"r"))) {
		fprintf(stderr,"%s: %s: %s\n",
			prgname,opt_file,strerror(errno));
		exit(DIALOG_ERROR);
	    }
	}
	
	if (!isatty(fileno(stdout))) {
	    int fd=dup(fileno(stdout));
	    outfile=fdopen(fd,"w");
	    opt_separate=1; /* always on different lines */
	    if (!freopen(tty,"w",stdout)) {
		fprintf(stderr,"%s: %s: %s\n",prgname,tty,strerror(errno));
		exit(DIALOG_ERROR);
	    }
	}
    }

    /*
     * Ok, now do any specific preprocessing for the dialog flavours.
     * This replaces the "jumperFn" stuff.
     */

    switch (desc.type) {
      case DIALOG_YESNO:
      case DIALOG_MSG:
      case DIALOG_INFO:

	/* nothing special to do */
	break; 

      case DIALOG_CHECK:
      case DIALOG_RADIO:
      case DIALOG_MENU:

	/* process extra args and build the list */
	{
	    int step=(desc.type==DIALOG_MENU ? 2 : 3);

	    if (opt_one) step=1;

	    /* get list_height */
	    if (argc>1 && sscanf(argv[1],"%hi%s",&desc.xtra,opt_opts)==1) {
		argv++,argc--;
	    }

	    argv++,argc--; /* drop the extra argument */

	    if (infile) {
		desc.clientdata=dialog_f2l(infile,step);
	    }
	    else {
		if (argc%step) {
		    fprintf(stderr,"%s: Wrong number of list args (%i) in "
			    "list-like box\n",prgname,argc);
		    exit(DIALOG_ERROR);
		}
		desc.clientdata=dialog_a2l(argc,argv,step);
	    }
	    if (!desc.clientdata) {
		fprintf(stderr,"%s: %s\n",prgname,strerror(errno));
		exit(DIALOG_ERROR);
	    }
	}
	break;
	
      case DIALOG_INPUT: /* process the last arg, as initial value */

	desc.clientdata=NULL;
	if (argc>1) {
	    desc.clientdata=argv[1];
	    argv++, argc--;
	}
	break;

      case DIALOG_GAUGE:
	if (argc>1 && sscanf(argv[1],"%hi%s",(&desc.xtra),opt_opts)==1) {
	    argv++,argc--;
	}
	break;

      case DIALOG_TEXT: /* nothing to do */
    }

    /*
     * And finally do your task
     */

    dialog_init ();
    retval = (*(dialogFn *)(optPtr->ptr))(&desc);

    /*
     * Print the result, if needed
     */
    if (retval==0)
      switch (desc.type) {
	case DIALOG_YESNO:
	case DIALOG_MSG:
	case DIALOG_INFO:
	case DIALOG_TEXT:
	case DIALOG_GAUGE:  /* nothing to do */
	  break;
	  
	case DIALOG_INPUT:
	  fprintf(outfile,dialog_input_result);
	  break;

	case DIALOG_CHECK:
	  for (lPtr=desc.clientdata; lPtr; lPtr=lPtr->next)
	    if (lPtr->selected==DIALOG_B_ENABLED)
	      fprintf (outfile, opt_separate ? "%s\n" : "\"%s\" ",
		       lPtr->tag);
	  break;

	case DIALOG_RADIO:
	case DIALOG_MENU:
	  for (lPtr=desc.clientdata; lPtr; lPtr=lPtr->next) {
	    if (lPtr->selected==DIALOG_B_ENABLED)
	      fprintf(outfile,"%s",lPtr->tag);}
/*
		      outfile==stderr ? "" : "\n");
*/
	  break;
      }
   
    if (opt_clear) dialog_clear ();
    refresh ();
    
    dialog_end ();
    exit (retval);
}

/*
 * Print program usage
 */
static void
Usage (const char *name)
{
    fprintf (stderr, "dialog version " DIALOG_VERSION ", \n"
 "Usage: %s [common options] <Box options>\n\n"
 " Box options:\n"
 "  --yesno     <text> <height> <width>\n"
 "  --msgbox    <text> <height> <width>\n"
 "  --infobox   <text> <height> <width>\n"
 "  --inputbox  <text> <height> <width> [<init>]\n"
 "  --textbox   <file> <height> <width>\n"
 "  --menu      <text> <hei> <wid> <menu_hei> <tag1> <item1> ...\n"
 "  --checklist <text> <hei> <wid> <list_hei> <tag1> <item1> <status1> ...\n"
 "  --radiolist <text> <hei> <wid> <list_hei> <tag1> <item1> <status1>...\n"
 "  --gauge     <text> <hei> <wid> <percent>\n",
	     name);
    exit (DIALOG_ERROR);
}



/**** Local Variables:  ****/
/**** tab-width:8       ****/
/**** c-indent-level:4  ****/
/**** c-label-offset:-2 ****/
/**** End:              ****/

