/*
 *  inputbox.c -- implements the input box
 *
 *  Originally by Savio Lam (lam836@cs.cuhk.hk)
 *  Completely reworked by Alessandro Rubini (rubini@ipvvis.unipv.it).
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "dialog.h"
#include "dialogInt.h"
#include <memory.h> /* memset, for SunOS */

char dialog_input_result[DIALOG_MAX_LEN + 1];

static char *c_b, *c_e, *c_c;      /* begin, end, current */
static char *c_bd, *c_ed, *c_tmp;  /* begin-display, end-display, tmp */

#define put_cursor()  wmove (dialog, box_y, box_x + (c_c - c_bd))

static void redraw_string(WINDOW *dialog, int box_y, int box_x)
{
    char tmp;

    wmove (dialog, box_y, box_x);
    wattrset (dialog, inputbox_attr);
    tmp=*c_ed; *c_ed='\0';
    waddstr (dialog, c_bd);
    *c_ed=tmp;
    put_cursor();
}

/*
 * Display a dialog box for inputing a (single) string
 */
int
dialog_inputbox (Dialog_desc *desc)
{
    int i, x, y, box_y, box_x, box_width, editing=1;
    int  key, button = desc->selected;
    int changed=0;
    WINDOW *dialog;
    struct Dialog_butt *bptr;
    char *instr = dialog_input_result;

    desc->hlinenum = 1;
    desc->hline[0] = -3;

    /* The caller can build the buttons beforehand, othwise, default applies */
    if (!desc->buttons)
      dialog_defaultbuttons(desc,"  OK  :Cancel");

    _dialog_setup(desc);
    dialog = desc->win;

    /*
     * Hmm.... the prompt has already been written, but the cursor
     * has already moved away, while drawing buttons. So I need to
     * save the cursor position, redraw the prompt, and restore the cursor.
     * This must be fixed when automagic positioning is implemented.
     */

    if (desc->prompt)
      _dialog_print_autowrap (dialog, desc->prompt, desc->wid - 2, 1, 3);

    /* horrible */
    x = (COLS - desc->wid) / 2;
    y = (LINES - desc->hei) / 2;

    /* Draw the input field box */
    getyx (dialog, y, x);
    box_y = y + 2;
    box_width = desc->wid - 6;
    box_x = 3;
    /*
     * use the '!' as mouse-hotkey for the input line 
     * (never used, and not documented ;-)
     */
    _dialog_mkregion (y + 1, box_x - 1, 3, box_width + 2, '!');
    _dialog_draw_box (dialog, y + 1, box_x - 1, 3, box_width + 2,
	      border_attr, dialog_attr);

    x = desc->wid / 2 - 11;
    y = desc->hei - 2;

    /* Set up pointers and fill the buffer with blanks */
    memset(instr,' ', sizeof(dialog_input_result));
    c_b=c_e=c_c=c_bd=c_ed=instr;

    if (desc->clientdata) { /* initial value -- move pointers */
	strcpy (instr, desc->clientdata);
	c_e+=strlen(instr);
	*c_e=' ';
	c_c=c_ed=c_e;
    }

    if (c_ed-c_bd > box_width)
      c_bd=c_ed-box_width;
    else
      c_ed=c_bd+box_width;

    redraw_string(dialog, box_y, box_x);
    wrefresh (dialog);

    while ( (key=_dialog_wgetch(dialog)) != DIALOG_ESC) {
 
	if (editing) {	/* Input box selected */
	    switch (key) {
	      case KEY_UP:
	      case KEY_DOWN:
	      case DIALOG_M_EVENT + '!':
		continue; /* ignore */

	      case '\n':   /* enter highlights "Ok", like TAB */
		key=DIALOG_TAB;

	      case DIALOG_TAB:
		editing=0;
		button=-1; /* tab is reprocessed, so button becomes 0 */
		break;

		break;

	      case 'A'-'@': /* control-A */
		c_c=c_b;
		break;

	      case 'E'-'@': /* control-E */
		c_c=c_e;
		break;

	      case 'K'-'@': /* control-K */
		changed++;
		memset(c_c,' ',c_e-c_c);
		c_e=c_c;
		break;

	      case 'B'-'@': /* control-B */
	      case KEY_LEFT:
		if (c_c > c_b)
		  c_c--;
		break;
	      case 'F'-'@': /* control-F */
	      case KEY_RIGHT:
		if (c_c < c_e)
		  c_c++;
		break;

	      case 'D'-'@': /* control-D */
		if (c_c==c_e) break;
		c_c++;
		/* fall through */

	      case KEY_BACKSPACE:
	      case 127:
		if (c_c == c_b)
		  break;
		changed++;
#if 0
	        memmove(c_c-1, c_c, c_e-c_c+1);
#else
		for (c_tmp=c_c-1; c_tmp<c_e; c_tmp++)
		  c_tmp[0]=c_tmp[1];
#endif
		c_c--; c_e--;
		break;

	      default:
		if (key > DIALOG_M_EVENT) {
		    editing=0;
		    break;
		}
		if (key < 0x100 && isprint (key)) { /* FIXME -- overflow */
		    changed++;
#if 0
		    memmove(c_c+1, c_c, c_e-c_c);
#else
		    for(c_tmp=c_e-1; c_tmp>=c_c; c_tmp--)
		      c_tmp[1]=c_tmp[0];
#endif
		    *c_c=key;
		    c_c++; c_e++;
		}
	    }

	    while (c_c < c_bd) {
		changed++;
		c_bd -= (box_width+1)/2;
		if (c_bd < c_b) c_bd=c_b;
		c_ed=c_bd+box_width;
	    }
	    while (c_c > c_ed) {
		changed++;
		c_bd+= (box_width+1)/2;
		c_ed=c_bd+box_width;
	    }
	      
	    if (changed) {
		changed=0;
		redraw_string(dialog, box_y, box_x);
	    }
	    else
	      put_cursor();
	    wrefresh(dialog);

	    if (editing)
	      continue;
	}

	/* not editing: it's like yesno.c */

	if (key < 0x100) key=tolower(key);

	/* first of all, look for an hotkey */
	for (i=0, bptr=desc->buttons; i<desc->buttnum; i++, bptr++) {
	    if ( key==(int)tolower(*bptr->hotkey) ||
		key==DIALOG_M_EVENT+(int)toupper(*bptr->hotkey) ) {
		delwin(dialog);
		*c_e='\0';
		return(i);
	    }
	    if (key == DIALOG_M_EVENT+(int)tolower(*bptr->hotkey)) {
		button = i;
		break;
	    }
	}

	switch (key) {

	  case DIALOG_TAB:
	    if (button == desc->buttnum-1)
	      editing=1;
	    if (button==-1) /* force a redraw by faking an inactive butt */
	      desc->buttons[0].status=DIALOG_B_DISABLED;

	    /* fall through */

	  case KEY_DOWN:   /* increment */
	  case KEY_RIGHT:
	    button++; button %=  desc->buttnum; break;
	    break;

	  case KEY_UP:
	  case KEY_LEFT:   /* decrement */
	    button--; if (button<0) button=desc->buttnum-1; break;

	  case ' ':
	  case '\n':
	    delwin (dialog);
	    *c_e='\0';
	    return button;
	}
	/* refresh only if needed */
	if (desc->buttons[button].status==DIALOG_B_DISABLED) {
	    desc->buttons[button].status=DIALOG_B_ENABLED;
	    _dialog_print_buttons(desc);
	    if(editing) put_cursor();
	    wrefresh (dialog);
	}
    }
    
    delwin (dialog);
    return -1;          /* ESC pressed */
}



/**** Local Variables:  ****/
/**** tab-width:8       ****/
/**** c-indent-level:4  ****/
/**** c-label-offset:-2 ****/
/**** End:              ****/

