/*
 *  util.c
 *
 *  Originally by Savio Lam (lam836@cs.cuhk.hk)
 *  Greatly upgraded by Alessandro Rubini (rubini@ipvvis.unipv.it).
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "dialog.h"
#include "dialogInt.h"
#include <errno.h>
#include <sys/ioctl.h>
#include <termios.h> /* winsize */


#ifdef HAVE_NCURSES
/* use colors by default? */
bool use_colors = DIALOG_USE_COLORS;
/* shadow dialog boxes by default?
   Note that 'use_shadow' implies 'use_colors' */
bool use_shadow = DIALOG_USE_SHADOW;
#endif

const char *dialog_result;

/* 
 * Attribute values, default is for mono display
 */
chtype attributes[] =
{
    A_NORMAL,			/* screen_attr */
    A_NORMAL,			/* shadow_attr */
    A_REVERSE,			/* dialog_attr */
    A_REVERSE,			/* title_attr */
    A_REVERSE,			/* border_attr */
    A_BOLD,			/* button_active_attr */
    A_DIM,			/* button_inactive_attr */
    A_UNDERLINE,		/* button_key_active_attr */
    A_UNDERLINE,		/* button_key_inactive_attr */
    A_NORMAL,			/* button_label_active_attr */
    A_NORMAL,			/* button_label_inactive_attr */
    A_REVERSE,			/* inputbox_attr */
    A_REVERSE,			/* inputbox_border_attr */
    A_REVERSE,			/* searchbox_attr */
    A_REVERSE,			/* searchbox_title_attr */
    A_REVERSE,			/* searchbox_border_attr */
    A_REVERSE,			/* position_indicator_attr */
    A_REVERSE,			/* menubox_attr */
    A_REVERSE,			/* menubox_border_attr */
    A_REVERSE,			/* item_attr */
    A_NORMAL,			/* item_selected_attr */
    A_REVERSE,			/* tag_attr */
    A_NORMAL,	/* was REV */	/* tag_selected_attr */
    A_NORMAL,			/* tag_key_attr */
    A_BOLD,			/* tag_key_selected_attr */
    A_REVERSE,			/* check_attr */
    A_REVERSE,			/* check_selected_attr */
    A_REVERSE,			/* uarrow_attr */
    A_REVERSE			/* darrow_attr */
};

#ifdef HAVE_NCURSES
#include "colors.h"

/*
 * Table of color values
 */
int color_table[][3] =
{
    {SCREEN_FG, SCREEN_BG, SCREEN_HL},
    {SHADOW_FG, SHADOW_BG, SHADOW_HL},
    {DIALOG_FG, DIALOG_BG, DIALOG_HL},
    {TITLE_FG, TITLE_BG, TITLE_HL},
    {BORDER_FG, BORDER_BG, BORDER_HL},
    {BUTTON_ACTIVE_FG, BUTTON_ACTIVE_BG, BUTTON_ACTIVE_HL},
    {BUTTON_INACTIVE_FG, BUTTON_INACTIVE_BG, BUTTON_INACTIVE_HL},
    {BUTTON_KEY_ACTIVE_FG, BUTTON_KEY_ACTIVE_BG, BUTTON_KEY_ACTIVE_HL},
    {BUTTON_KEY_INACTIVE_FG, BUTTON_KEY_INACTIVE_BG, BUTTON_KEY_INACTIVE_HL},
    {BUTTON_LABEL_ACTIVE_FG, BUTTON_LABEL_ACTIVE_BG, BUTTON_LABEL_ACTIVE_HL},
    {BUTTON_LABEL_INACTIVE_FG, BUTTON_LABEL_INACTIVE_BG,
     BUTTON_LABEL_INACTIVE_HL},
    {INPUTBOX_FG, INPUTBOX_BG, INPUTBOX_HL},
    {INPUTBOX_BORDER_FG, INPUTBOX_BORDER_BG, INPUTBOX_BORDER_HL},
    {SEARCHBOX_FG, SEARCHBOX_BG, SEARCHBOX_HL},
    {SEARCHBOX_TITLE_FG, SEARCHBOX_TITLE_BG, SEARCHBOX_TITLE_HL},
    {SEARCHBOX_BORDER_FG, SEARCHBOX_BORDER_BG, SEARCHBOX_BORDER_HL},
    {POSITION_INDICATOR_FG, POSITION_INDICATOR_BG, POSITION_INDICATOR_HL},
    {MENUBOX_FG, MENUBOX_BG, MENUBOX_HL},
    {MENUBOX_BORDER_FG, MENUBOX_BORDER_BG, MENUBOX_BORDER_HL},
    {ITEM_FG, ITEM_BG, ITEM_HL},
    {ITEM_SELECTED_FG, ITEM_SELECTED_BG, ITEM_SELECTED_HL},
    {TAG_FG, TAG_BG, TAG_HL},
    {TAG_SELECTED_FG, TAG_SELECTED_BG, TAG_SELECTED_HL},
    {TAG_KEY_FG, TAG_KEY_BG, TAG_KEY_HL},
    {TAG_KEY_SELECTED_FG, TAG_KEY_SELECTED_BG, TAG_KEY_SELECTED_HL},
    {CHECK_FG, CHECK_BG, CHECK_HL},
    {CHECK_SELECTED_FG, CHECK_SELECTED_BG, CHECK_SELECTED_HL},
    {UARROW_FG, UARROW_BG, UARROW_HL},
    {DARROW_FG, DARROW_BG, DARROW_HL},
};				/* color_table */
#endif

/*
 * Set window to attribute 'attr'
 */
void
_dialog_attr_clear (WINDOW * win, int height, int width, chtype attr)
{
    int i, j;

    wattrset (win, attr);
    for (i = 0; i < height; i++) {
	wmove (win, i, 0);
	for (j = 0; j < width; j++)
	    waddch (win, ' ');
    }
    touchwin (win);
}

void dialog_clear (void)
{
    _dialog_attr_clear (stdscr, LINES, COLS, screen_attr);
    wnoutrefresh (stdscr);
}

#ifdef HAVE_NCURSES
/*
 * Setup for color display
 */
static void
color_setup (void)
{
    int i;

    if (has_colors ()) {	/* Terminal supports color? */
	start_color ();

	/* Initialize color pairs */
	for (i = 0; i < ATTRIBUTE_COUNT; i++)
	    init_pair (i + 1, color_table[i][0], color_table[i][1]);

	/* Setup color attributes */
	for (i = 0; i < ATTRIBUTE_COUNT; i++)
	    attributes[i] = C_ATTR (color_table[i][2], i + 1);
    }
}
#endif

/*
 * Do some initialization for dialog
 */
void
dialog_init (void)
{
    _dialog_m_open (); /* before curses, to avoid races with curses mouse */

#ifndef CFG_NORCFILE /* read rc unless specifically disabled */
#  ifdef HAVE_NCURSES
    if (_dialog_parse_rc () == -1)
	exit (DIALOG_ERROR);
#  endif
#endif

    initscr ();			/* Init curses */
    keypad (stdscr, TRUE);
    cbreak ();
    noecho ();

#ifdef HAVE_NCURSES
    if (use_colors || use_shadow)	/* Set up colors */
	color_setup ();
#endif

    dialog_clear ();
}

/*
 * End using dialog functions.
 */
void
dialog_end (void)
{
    _dialog_m_close ();
    _dialog_m_cleanup (); /* just in case... */
    endwin ();
}

/*
 * Common code to setup it all. Overrides the previous code.
 * Returns int 'cause returning void is silly. An error can always happen.
 */
int
_dialog_setup (struct Dialog_desc *desc)
{
    struct winsize ws;
    WINDOW *win;
    int x,y; /* ul corner */


    if (!desc->win) { /* if win is not NULL, this is only an update */

	/*
	 * First: window size. Check either stdout or stdin or open /dev/tty.
	 * In case of failure should fail altogether, but since curses wants
	 * to go on (its fault), default to 80x25. I'm scary in using LINES
	 * and COLS. I have to check about curses and size changes.
	 */
	{
	    int ttyfd=-1;
	    int needclose=0;
	    
	    if (isatty(fileno(stdout))) ttyfd=fileno(stdout);
	    else if (isatty(fileno(stdin))) ttyfd=fileno(stdin);
	    else {
		ttyfd=open("/dev/tty",O_RDONLY);
		needclose=1;
	    }
	    ws.ws_row=25; ws.ws_col=80; 
	    if (ttyfd>=0)
	      ioctl(ttyfd, TIOCGWINSZ, &ws); /* if failed, default applies */
	    if (needclose)
	      close(ttyfd);
	    if (ws.ws_row==0) { ws.ws_row=25; ws.ws_col=80; } /* default */
	}
	
	/*
	 * Ok, now fix width and height. Default to a margin of 3 on each side.
	 */
	if (desc->c_wid==0) desc->c_wid=-3;
	if (desc->c_hei==0) desc->c_hei=-3;
	
	desc->wid=desc->c_wid; desc->hei=desc->c_hei;
	
	if (desc->wid<0) desc->wid = ws.ws_col+2*desc->wid;
	if (desc->hei<0) desc->hei = ws.ws_row+2*desc->hei;
	
	x = (ws.ws_col-desc->wid) / 2;
	y = (ws.ws_row-desc->hei) / 2;
#if 1
	if (desc->backtitle != NULL) {
	    int i;
	    
	    wattrset (stdscr, screen_attr);
	    wmove (stdscr, 0, 1);
	    waddstr (stdscr, desc->backtitle);
	    wmove (stdscr, 1, 1);
	    for (i = 1; i < COLS - 1; i++)
	      waddch (stdscr, ACS_HLINE);
	}
#endif
	/*
	 * The following replaces _dialog_newwin().
	 */
#ifdef HAVE_NCURSES
	if (use_shadow)
	  _dialog_draw_shadow (stdscr, y, x, desc->hei, desc->wid);
#endif
	win = desc->win = newwin (desc->hei, desc->wid, y, x);
	_dialog_setbase (x, y);
	keypad (win, TRUE);
    }
    else { /* redraw */
	win=desc->win;
	werase(desc->win);
    }

    /* always do the following, even if redrawing */
	_dialog_draw_box (win, 0, 0, desc->hei, desc->wid, dialog_attr,
			  border_attr);
	
    /*
     * The following replaces _dialog_title()
     */
    if (desc->title) {
        wattrset (win, title_attr);
	wmove (win, 0, (desc->wid - strlen (desc->title)) / 2 - 1);
	waddch (win, ' ');
	waddstr (win, desc->title);
	waddch (win, ' ');
    }
    
    /*
     * Then, draw horizontal lines, if any.
     */
    {
        int i,j;
	for (i=0; i<desc->hlinenum; i++) {
            int vpos = desc->hline[i];
	    wattrset (win, border_attr);
	    wmove (win, vpos<0 ? desc->hei+vpos : vpos, 0);
	    waddch (win, ACS_LTEE);
	    for (j = 0; j < desc->wid-2; j++)
	        waddch (win, ACS_HLINE);
	    wattrset (win, dialog_attr);
	    waddch (win, ACS_RTEE);
	}
    }

    /*
     * Draw the prompt...
     */

    if (desc->prompt)
      _dialog_print_autowrap (win, desc->prompt, desc->wid - 2, 1, 3);

    /*
     * And finally, draw the buttons.
     */
    if (desc->buttons)
      _dialog_print_buttons(desc);

    return 0; /* ok */
}
   

/*
 * Print a string of text in a window, automatically wrap around to the
 * next line if the string is too long to fit on one line. Note that the
 * string may contain "\n" to represent a newline character or the real
 * newline '\n', but in that case, auto wrap around will be disabled.
 */
void
_dialog_print_autowrap (WINDOW * win, const char *prompt,
			int width, int y, int x)
{
    int first = 1, cur_x, cur_y;
    char tempstr[DIALOG_MAX_LEN + 1], *word, *tempptr, *tempptr1;

    wattrset (win, dialog_attr);
    strcpy (tempstr, prompt);
    if ((strstr (tempstr, "\\n") != NULL) || (strchr (tempstr, '\n') != NULL)) {
	/* Prompt contains "\n" or '\n' */
	word = tempstr;
	cur_y = y;
	wmove (win, cur_y, x);
	while (1) {
	    tempptr = strstr (word, "\\n");
	    tempptr1 = strchr (word, '\n');
	    if (tempptr == NULL && tempptr1 == NULL)
		break;
	    else if (tempptr == NULL) {		/* No more "\n" */
		tempptr = tempptr1;
		tempptr[0] = '\0';
	    } else if (tempptr1 == NULL) {	/* No more '\n' */
		tempptr[0] = '\0';
		tempptr++;
	    } else {		/* Prompt contains both "\n" and '\n' */
		if (strlen (tempptr) - 2 < strlen (tempptr1) - 1) {
		    tempptr = tempptr1;
		    tempptr[0] = '\0';
		} else {
		    tempptr[0] = '\0';
		    tempptr++;
		}
	    }

	    waddstr (win, word);
	    word = tempptr + 1;
	    wmove (win, ++cur_y, x);
	}
	waddstr (win, word);
    } else if (strlen (tempstr) <= width - x * 2) {	/* If prompt is short */
	wmove (win, y, (width - strlen (tempstr)) / 2);
	waddstr (win, tempstr);
    } else {
	cur_x = x;
	cur_y = y;
	/* Print prompt word by word, wrap around if necessary */
	while ((word = strtok (first ? tempstr : NULL, " ")) != NULL) {
	    if (first)		/* First iteration */
		first = 0;
	    if (cur_x + strlen (word) > width) {
		cur_y++;	/* wrap to next line */
		cur_x = x;
	    }
	    wmove (win, cur_y, cur_x);
	    waddstr (win, word);
	    getyx (win, cur_y, cur_x);
	    cur_x++;
	}
    }
}


/*
 * The new routine. Static because it is only used by the following one.
 */

static int
_dialog_button (WINDOW *win, struct Dialog_butt *button)
{
    int selected= (button->status==DIALOG_B_CHOSEN);
    char *label=button->text;

    wmove (win, button->y, button->x);
    wattrset (win, selected ? button_active_attr : button_inactive_attr);
    waddstr (win, "<");

    if (!button->moused) {
	_dialog_mkbutton (button->y, button->x, strlen(button->text)+2,
			tolower(*(button->hotkey)));
	button->moused++;
    }
			
    wattrset (win, selected ? button_label_active_attr
              : button_label_inactive_attr);
    while (label< button->hotkey && *label) {
        waddch (win, *label);
	label++;
    }
    wattrset (win, selected ? button_key_active_attr
              : button_key_inactive_attr);
    if (*label) waddch (win, *label); {
	wattrset (win, selected ? button_label_active_attr
		  : button_label_inactive_attr);
	waddstr (win, label + 1);
    }
    wattrset (win, selected ? button_active_attr : button_inactive_attr);
    waddstr (win, ">");
    /* by printing them in order, it would always remain on the last butt */
    wmove (win, 0,0); 
    return 0; /* ok */
}

int
_dialog_print_buttons (struct Dialog_desc *desc)
{
    WINDOW *win=desc->win;
    struct Dialog_butt *buttons=desc->buttons;
    int i,j;

    if (!buttons)
      return 0;

    /*
     * First of all, choose the positions, if missing
     */
    if (!buttons->x) {   /* horiz */
        int totwid=0, left;
	for (i=0; i<desc->buttnum; i++)
	  totwid+=strlen(buttons[i].text)+2;
        left=desc->wid - 2 - totwid;
	left /= 1+desc->buttnum; /* FIXME -- Aestethic issues */
	j=left+1;
	for (i=0; i<desc->buttnum; i++) {
	    buttons[i].x=j;
	    j += strlen(buttons[i].text)+2+left;
	}
    }
    if (!buttons->y)
        for (i=0; i<desc->buttnum; i++)
	    buttons[i].y=desc->hei-2;

    /*
     * Then, print them (check the enabled one)
     */
    for (i=0; i<desc->buttnum; i++)
        if (buttons[i].status==DIALOG_B_ENABLED)
	    break;
    if (i<desc->buttnum)
        buttons[i].status=DIALOG_B_CHOSEN;
    else
        for (i=desc->buttnum; --i>0; )
	    if (buttons[i].status==DIALOG_B_CHOSEN)
	        break;

    /*
     * Now, print. The current "i" is the only enabled one
     */
    for (j=0;j<desc->buttnum; j++) {
	buttons[j].status= (j==i ? DIALOG_B_CHOSEN : DIALOG_B_DISABLED);
        _dialog_button(win,buttons+j);
    }

    /*
     * Finally, move the cursor to the hotkey, or to the button
     */
    buttons+=i; /* use it */
    wmove(win,buttons->y, buttons->x +
	  (buttons->hotkey ? buttons->hotkey-buttons->text+1 : 1) );

    return 0;
}

/*
 * Create buttons, with specified text in them
 */
int
dialog_defaultbuttons(struct Dialog_desc *desc, char *opts)
{
  int n=1;
  char *s;
  char *text=strdup(opts); /* constant strings are not writable */
  struct Dialog_butt *button;

  for (s=opts; *s; s++)
    n+= (*s==':');

  button=calloc(n,sizeof(struct Dialog_butt));
  if (!button) return -1;
  desc->buttons=button;
  desc->buttnum=n;

  button->text=text;
  while (*text) {
      if (*text==':') {
          *text='\0';
	  (++button)->text=++text;
      }
      else
	  ++text;
  }
  /*
   * Create the hotkeys: the first uppercase, or the first letter
   */
  for (n=0, button=desc->buttons; n<desc->buttnum; n++, button++) {
      for (s=button->text; *s && *s==' '; s++)
	;
      button->hotkey= (*s ? s : button->text);
      while (*s && !isupper(*s)) s++;
      if (*s) button->hotkey=s;
  }

  /*
   * And select the chosen one
   */
  if (desc->selected < desc->buttnum)
    desc->buttons[desc->selected].status=DIALOG_B_ENABLED;

  return 0;
}

/*
 * Draw a rectangular box with line drawing characters
 */
void
_dialog_draw_box (WINDOW * win, int y, int x, int height, int width,
	  chtype box, chtype border)
{
    int i, j;

    wattrset (win, 0);
    for (i = 0; i < height; i++) {
	wmove (win, y + i, x);
	for (j = 0; j < width; j++)
	    if (!i && !j)
		waddch (win, border | ACS_ULCORNER);
	    else if (i == height - 1 && !j)
		waddch (win, border | ACS_LLCORNER);
	    else if (!i && j == width - 1)
		waddch (win, box | ACS_URCORNER);
	    else if (i == height - 1 && j == width - 1)
		waddch (win, box | ACS_LRCORNER);
	    else if (!i)
		waddch (win, border | ACS_HLINE);
	    else if (i == height - 1)
		waddch (win, box | ACS_HLINE);
	    else if (!j)
		waddch (win, border | ACS_VLINE);
	    else if (j == width - 1)
		waddch (win, box | ACS_VLINE);
	    else
		waddch (win, box | ' ');
    }
}

#ifdef HAVE_NCURSES
/*
 * Draw shadows along the right and bottom edge to give a more 3D look
 * to the boxes
 */
void
_dialog_draw_shadow (WINDOW * win, int y, int x, int height, int width)
{
    int i;

    if (has_colors ()) {	/* Whether terminal supports color? */
	wattrset (win, shadow_attr);
	wmove (win, y + height, x + 2);
	for (i = 0; i < width; i++)
	  waddch (win, winch (win) & A_CHARTEXT);
	for (i = y + 1; i < y + height + 1; i++) {
	    wmove (win, i, x + width);
	    waddch (win, winch (win) & A_CHARTEXT);
	    waddch (win, winch (win) & A_CHARTEXT);
	}
	wnoutrefresh (win);
    }
}
#endif

/*
 * Convert an array to a list. Step==1: only tag. Step==2: tag+text.
 * Step==3: tag+text+status
 */

Dialog_list *
dialog_a2l(int argc, char **argv, int step)
{
    Dialog_list *list, *curr;
    int n, i;
    char *hotstring, *s;

    if (argc%step) return NULL;
    if (step<1 || step>3) return NULL;

    n=argc/step;
    list=calloc(n,sizeof(Dialog_list));
    if (!list) return NULL;

    /* now fill the array like it was a list */
    for (curr=list, i=0; i<n; i++,curr++) {
	if (i)       curr->prev = curr-1;
	if (i < n-1) curr->next = curr+1;
	curr->index=i;
	curr->tag = *(argv++);
	hotstring = curr->tag;
	if (step>1)  {
	    curr->text = *(argv++);
	}
	if (step>2)  {
	    if (!strcasecmp(*argv,"on")) curr->selected=DIALOG_B_ENABLED;
	    else if (!strcasecmp(*argv,"off")) /* nothing */;
	    else if (**argv=='0' || tolower(**argv)=='n') /* nothing */;
	    else curr->selected=DIALOG_B_ENABLED;
	    argv++;
	}
	/* look for the hot key */
	for (s=hotstring; *s && *s==' '; s++)
	  ;
	curr->hotkey= (*s ? s : hotstring);
	while (*s && !isupper(*s)) s++;
	if (*s) curr->hotkey=s;
    }

#if 0
	fprintf(stderr,"txt \"%s\", tag \"%s\", sel %i\n",
		curr->text,curr->tag,curr-> selected);
#endif

    return list;
}

/*
 * Do the same for a file.
 */

Dialog_list *
dialog_f2l(FILE *f, int step)
{
    char **argv;
    int argc=0;
    Dialog_list *l=NULL;
    char s[128], *t;
    int saverrno;

    argv=malloc(1024*sizeof(char *));
    if (!argv) return NULL;

    /* build an array, and call dialog_a2l() */
    while (fgets(s,128,f) && argc<1024) {
	if (s[0]=='#') continue; /* comment */
	t=s;
	if (s[0]=='\\') t++; /* skip the backslash */
	if (s[strlen(s)-1]=='\n') /* drop the newline */
	  s[strlen(s)-1]='\0';
	argv[argc++]=strdup(t);
	if (!argv[argc-1]) break; /* free() is missing */
    }

    if (argv[argc-1]) l=dialog_a2l(argc,argv,step);
    saverrno=errno;
    free(argv);
    errno=saverrno;
    return l;
}

/*
 * Run a dialog by name and description
 */
int dialog_run(int flavour, Dialog_desc *desc)
{
    switch( (desc->type=flavour) ) {
      case DIALOG_YESNO:
      case DIALOG_MSG:
      case DIALOG_INFO:
	return dialog_simple(desc);
      case DIALOG_CHECK:
      case DIALOG_RADIO:
      case DIALOG_MENU:
	return dialog_list(desc);
      case DIALOG_INPUT:
	return dialog_inputbox(desc);
      case DIALOG_TEXT:
	return dialog_textbox(desc);
      case DIALOG_GAUGE:
	return dialog_gauge(desc);
      default:
	return DIALOG_ERROR;
    }
}




/**** Local Variables:  ****/
/**** tab-width:8       ****/
/**** c-indent-level:4  ****/
/**** c-label-offset:-2 ****/
/**** End:              ****/



